<?php
// This file is part of Multi-tenancy plugin for Moodle™.

namespace tool_mutenancy\phpunit\external;

use tool_mutenancy\external\form_associate_add_userids;

/**
 * Multi-tenancy external function tests.
 *
 * @group       muTMS
 * @package     tool_mutenancy
 * @copyright   2025 Petr Skoda
 * @license     https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 *
 * @coversDefaultClass \tool_mutenancy\external\form_associate_add_userids
 */
final class form_associate_add_userids_test extends \advanced_testcase {
    public function setUp(): void {
        $this->resetAfterTest();
    }

    /**
     * @covers ::execute
     */
    public function test_execute(): void {
        /** @var \tool_mutenancy_generator $generator */
        $generator = $this->getDataGenerator()->get_plugin_generator('tool_mutenancy');

        $syscontext = \context_system::instance();

        $roleid = create_role('man', 'man', 'man');
        assign_capability('tool/mutenancy:view', CAP_ALLOW, $roleid, $syscontext->id);
        assign_capability('moodle/cohort:assign', CAP_ALLOW, $roleid, $syscontext->id);

        $tenant1 = $generator->create_tenant();
        $tenant2 = $generator->create_tenant();
        $tenant3 = $generator->create_tenant();

        $tenantcontext1 = \context_tenant::instance($tenant1->id);
        $categorycontext = \context_coursecat::instance($tenant1->categoryid);

        $cohort1 = $this->getDataGenerator()->create_cohort(['contextid' => $categorycontext->id]);
        $tenant1->assoccohortid = $cohort1->id;
        $tenant1 = \tool_mutenancy\local\tenant::update($tenant1);

        $cohort2 = $this->getDataGenerator()->create_cohort(['contextid' => $categorycontext->id, 'component' => 'core_bc']);
        $tenant2->assoccohortid = $cohort2->id;
        $tenant2 = \tool_mutenancy\local\tenant::update($tenant2);

        $admin = get_admin();
        $manager = $this->getDataGenerator()->create_user([
            'firstname' => 'Global',
            'lastname' => 'Manager',
            'email' => 'manager@example.com',
        ]);
        role_assign($roleid, $manager->id, $categorycontext->id);
        role_assign($roleid, $manager->id, $tenantcontext1->id);

        $user0 = $this->getDataGenerator()->create_user([
            'firstname' => 'Global',
            'lastname' => 'User',
            'email' => 'user0@example.com',
        ]);
        $user1 = $this->getDataGenerator()->create_user([
            'firstname' => 'First',
            'lastname' => 'User',
            'email' => 'user1@example.com',
        ]);
        $user2 = $this->getDataGenerator()->create_user([
            'firstname' => 'Second',
            'lastname' => 'User',
            'email' => 'user2@example.com',
        ]);
        $user3 = $this->getDataGenerator()->create_user([
            'tenantid' => $tenant1->id,
            'firstname' => 'Third',
            'lastname' => 'User',
            'email' => 'user3@example.com',
        ]);

        cohort_add_member($cohort1->id, $user2->id);

        $this->setUser($manager);
        $result = form_associate_add_userids::execute('', $tenant1->id);
        $this->assertNull($result['notice']);
        $this->assertCount(4, $result['list']);
        $this->assertSame($result['list'][0]['value'], $manager->id);
        $this->assertSame($result['list'][1]['value'], $admin->id);
        $this->assertSame($result['list'][2]['value'], $user1->id);
        $this->assertSame($result['list'][3]['value'], $user0->id);

        $this->setUser($manager);
        $result = form_associate_add_userids::execute('First', $tenant1->id);
        $this->assertNull($result['notice']);
        $this->assertCount(1, $result['list']);
        $this->assertSame($result['list'][0]['value'], $user1->id);

        $this->setUser($admin);
        $result = form_associate_add_userids::execute('First', $tenant1->id);
        $this->assertNull($result['notice']);
        $this->assertCount(1, $result['list']);
        $this->assertSame($result['list'][0]['value'], $user1->id);

        $this->setUser($manager);
        try {
            form_associate_add_userids::execute('', $tenant2->id);
            $this->fail('Exception expected');
        } catch (\core\exception\moodle_exception $ex) {
            $this->assertInstanceOf(\required_capability_exception::class, $ex);
            $this->assertSame('Sorry, but you do not currently have permissions to do that (View tenant).', $ex->getMessage());
        }

        $this->setUser($admin);
        try {
            form_associate_add_userids::execute('', $tenant2->id);
            $this->fail('Exception expected');
        } catch (\core\exception\moodle_exception $ex) {
            $this->assertInstanceOf(\invalid_parameter_exception::class, $ex);
            $this->assertSame('Invalid parameter value detected (Associate cohort cannot belong to any component)', $ex->getMessage());
        }

        $this->setUser($admin);
        try {
            form_associate_add_userids::execute('', $tenant3->id);
            $this->fail('Exception expected');
        } catch (\core\exception\moodle_exception $ex) {
            $this->assertInstanceOf(\invalid_parameter_exception::class, $ex);
            $this->assertSame('Invalid parameter value detected (tenant does not have associated cohort)', $ex->getMessage());
        }
    }

    /**
     * @covers ::get_label_callback
     */
    public function test_get_label_callback(): void {
        /** @var \tool_mutenancy_generator $generator */
        $generator = $this->getDataGenerator()->get_plugin_generator('tool_mutenancy');

        $cohort1 = $this->getDataGenerator()->create_cohort();
        $tenant1 = $generator->create_tenant(['assoccohortid' => $cohort1->id]);

        $user0 = $this->getDataGenerator()->create_user([
            'firstname' => 'Global',
            'lastname' => 'User',
            'email' => 'user0@example.com',
        ]);
        $user1 = $this->getDataGenerator()->create_user([
            'tenantid' => $tenant1->id,
            'firstname' => 'First',
            'lastname' => 'User',
            'email' => 'user1@example.com',
        ]);

        $callback = form_associate_add_userids::get_label_callback(['tenantid' => $tenant1->id]);

        $result = $callback($user0->id);
        $this->assertStringContainsString('Global User', $result);
        $this->assertStringNotContainsString($user0->email, $result);

        $this->setAdminUser();
        $result = $callback($user0->id);
        $this->assertStringContainsString('Global User', $result);
        $this->assertStringContainsString($user0->email, $result);
    }

    /**
     * @covers ::validate_userid
     */
    public function test_validate_userid(): void {
        /** @var \tool_mutenancy_generator $generator */
        $generator = $this->getDataGenerator()->get_plugin_generator('tool_mutenancy');

        $cohort1 = $this->getDataGenerator()->create_cohort();
        $tenant1 = $generator->create_tenant(['assoccohortid' => $cohort1->id]);

        $user0 = $this->getDataGenerator()->create_user([
            'firstname' => 'Global',
            'lastname' => 'User',
            'email' => 'user0@example.com',
        ]);
        $user1 = $this->getDataGenerator()->create_user([
            'tenantid' => $tenant1->id,
            'firstname' => 'First',
            'lastname' => 'User',
            'email' => 'user1@example.com',
        ]);

        $this->assertSame(null, form_associate_add_userids::validate_userid($user0->id, $tenant1->id));
        $this->assertSame('Error', form_associate_add_userids::validate_userid($user1->id, $tenant1->id));
    }
}
